--
-- Real Mower for FS19
-- @author:    	kenny456 (kenny456@seznam.cz)
-- @history:	v1.0 - 2020-06-30 - first release
--
RealMower = {};
RealMower.confDir = getUserProfileAppPath().. "modsSettings/RealMower/";
RealMower.modDirectory = g_currentModDirectory

function RealMower.prerequisitesPresent(specializations)
	return true
end
function RealMower.registerOverwrittenFunctions(vehicleType)
end

function RealMower.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "groundRaycastCallback", 				RealMower.groundRaycastCallback)
	SpecializationUtil.registerFunction(vehicleType, "toggleActive", 						RealMower.toggleActive)
end
function RealMower:onRegisterActionEvents(isSelected, isOnActiveVehicle)
	local spec = self.spec_realMower
	if not spec.modInitialized then
		return
	end
	
	if g_dedicatedServerInfo ~= nil then
		return
	end
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	end
	if self:getIsActiveForInput() then
		local actions = { InputAction.REALMOWER_ACTIVATE }

		for _,actionName in pairs(actions) do
			local always = (actionName == InputAction.REALSHOVEL_MENU_PLUS or actionName == InputAction.REALSHOVEL_MENU_MINUS) and true or false
			local _, eventID = g_inputBinding:registerActionEvent(actionName, self, RealMower.actionCallback, true, true, always, true)
			spec.event_IDs[actionName] = eventID
			if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
				if actionName == 'something with lower priority' then
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_NORMAL)
				else
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_VERY_HIGH)
				end
				if actionName == InputAction.REALMOWER_TOGGLE_HELP then
					g_inputBinding:setActionEventTextVisibility(eventID, true)
				elseif actionName == InputAction.REALMOWER_TOGGLE then
					g_inputBinding:setActionEventTextVisibility(eventID, false)
				else
					g_inputBinding:setActionEventTextVisibility(eventID, RealMower.showHelp)
				end
			end
			local colliding = false
			_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
			if colliding then
				if g_inputBinding.nameActions[actionName].bindings[1] ~= nil then
					if g_inputBinding.nameActions[actionName].bindings[1].inputString ~= nil then
						print(string.format('Warning: RealMower got a colliding input action: %s', actionName)..' ('..g_inputBinding.nameActions[actionName].bindings[1].inputString..'). You can remap it in controls settings')
					end
				else
					print(string.format('Warning: RealMower got a colliding input action: %s', actionName))
				end
			end
		end
	end
end
function RealMower.registerEventListeners(vehicleType)
	for _,n in pairs( { "onLoad", "onPostLoad", "onUpdate", "onRegisterActionEvents", "toggleActive", "onReadStream", "onWriteStream"} ) do
		SpecializationUtil.registerEventListener(vehicleType, n, RealMower)
	end
end
function RealMower:onLoad(savegame)
	self.spec_realMower = {}
	local spec = self.spec_realMower
	
	spec.modInitialized = false
	spec.realWorkAreas = {}
	if self.spec_workArea.workAreas ~= nil then
		for _, workArea in pairs(self.spec_workArea.workAreas) do
			if workArea.type == WorkAreaType.MOWER then
				spec.modInitialized = true
				local wx,wy,wz = getWorldTranslation(workArea.start)
				local a,b,c = getWorldTranslation(workArea.width)
				local wa,wb,wc = getWorldTranslation(workArea.height)
				local workWidth = MathUtil.vector3Length(wx - a, wy - b, wz - c)
				local workHeight = math.min(MathUtil.vector3Length(wx - wa, wy - wb, wz - wc), MathUtil.vector3Length(a - wa, b - wb, c - wc))
				workArea.startReal = {}
				workArea.widthReal = {}
				workArea.heightReal = {}
				workArea.lastStartWorldX = {}
				workArea.lastStartWorldZ = {}
				local j = 0
				local k = 0.62
				local k = 0.61
				local xk = 0
				local zk = 0
				local x,y,z = getTranslation(workArea.start)
				local a,b,c = getTranslation(workArea.width)
				local l,m,n = getTranslation(workArea.height)
				if x < a then
					xk = 1
					if a < l then
						workWidth = workWidth + (math.abs(l) - math.abs(a))
					end
				else
					xk = -1
					if a > l then
						workWidth = workWidth + (math.abs(l) - math.abs(a))
					end
				end
				if z > n then
					zk = -1
				else
					zk = 1
				end
				k = k * xk
				local number = math.ceil(workWidth / math.abs(k))
				local number = math.floor(workWidth / math.abs(k))
				local realWorkWidth = number * math.abs(k)
				if realWorkWidth > workWidth then
					j = (realWorkWidth - workWidth) / 2 * xk * -1
				elseif realWorkWidth < workWidth then
					j = (workWidth - realWorkWidth) / 2 * xk * 1
				else
					j = 0
				end
				local tempZ = 0
				local parent = getParent(getParent(workArea.start))
				local parentReal = createTransformGroup("parentReal")
				link(parent, parentReal)
				setTranslation(parentReal, getTranslation(getParent(workArea.start)))
				setRotation(parentReal, getRotation(getParent(workArea.start)))
				local x,y,z = worldToLocal(parentReal, wx, wy, wz)
				for i = 1, number do
					workArea.startReal[i] = createTransformGroup("startReal")
					link(parentReal, workArea.startReal[i])
					setTranslation(workArea.startReal[i], x+j,y+0.2,z+(workHeight/2)*zk + tempZ)
					workArea.widthReal[i] = createTransformGroup("widthReal")
					link(parentReal, workArea.widthReal[i])
					setTranslation(workArea.widthReal[i], x+j+k,y+0.2,z+(workHeight/2)*zk + tempZ)
					workArea.heightReal[i] = createTransformGroup("heightReal")
					link(parentReal, workArea.heightReal[i])
					setTranslation(workArea.heightReal[i], x+j,y+0.2,z+(workHeight/2+0.1)*zk + tempZ)
					j = j + k
					workArea.lastStartWorldX[i] = -4000
					workArea.lastStartWorldZ[i] = -4000
					workArea.totalArea = 0
					workArea.totalModedArea = 0
					workArea.totalRemoved = 0
					workArea.totalNotRemoved = 0
					workArea.infoText = ''
					workArea.infoTextTemp = ''
				end
				table.insert(spec.realWorkAreas, workArea)
			end
		end
	end
	if not spec.modInitialized then
		print("Error: RealMower initialization failed for "..tostring(self:getFullName()).." !")
		return
	end
	local xmlFile = self.xmlFile
	spec.event_IDs = {}
	spec.modActive = true
	RealMower.showHelp = true
	spec.groundRaycastResult = {}
    spec.groundRaycastResult.y = 0.0
    spec.groundRaycastResult.object = nil
    spec.groundRaycastResult.distance = 0.0
	
	if g_currentMission.densityMapModifiers['updateRealMowerArea'] == nil then
        local dMM = {}
        dMM.modifier = DensityMapModifier:new(g_currentMission.terrainDetailId,0,3)
        dMM.filter = DensityMapFilter:new(dMM.modifier)
        dMM.mask = DensityMapFilter:new(g_currentMission.terrainDetailId,0,3)
        g_currentMission.densityMapModifiers['updateRealMowerArea'] = dMM
    end
end
function RealMower:onPostLoad(savegame)
	local spec = self.spec_realMower
	if not spec.modInitialized then
		return
	end
	
	local xmlFile = self.xmlFile
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	end
	if savegame ~= nil then
		local xmlFile = savegame.xmlFile
		local key = savegame.key ..".RealMower"
		spec.modActive = Utils.getNoNil(getXMLBool(xmlFile, key.."#modActive"), spec.modActive)
		self:toggleActive(spec.modActive, true)
	end
end
function RealMower:saveToXMLFile(xmlFile, key)
	local spec = self.spec_realMower
	
	setXMLBool(xmlFile, key.."#modActive", spec.modActive)
end
function RealMower:onDelete()
	local spec = self.spec_realMower
	if not spec.modInitialized then
		return
	end
end
function RealMower:groundRaycastCallback(hitObjectId, x, y, z, distance)
    local spec = self.spec_realMower
    spec.groundRaycastResult.y = y
    spec.groundRaycastResult.object = hitObjectId
    spec.groundRaycastResult.distance = distance
    return false
end
function RealMower:onUpdate(dt, vehicle)
	local spec = self.spec_realMower
	
	if not spec.modInitialized then
		return
	end
	spec.infoText = ''
	if self:getIsActive() and spec.modActive then
		for n, workArea in pairs(spec.realWorkAreas) do
			workArea.infoText = ''
			local number = #workArea.startReal
			local half = math.ceil(number / 2)
			local x6,_,z6 = getWorldTranslation(workArea.startReal[half])
			local diff = MathUtil.vector2Length(x6-workArea.lastStartWorldX[half], z6-workArea.lastStartWorldZ[half])
			local kon = 0.6
			if diff > kon and (self:getIsWorkAreaActive(self.spec_workArea.workAreas[n]) and g_currentMission.accessHandler:canFarmAccessLand(self:getActiveFarm(), x6, z6)) then
				local changed = {}
				for i, area in pairs(workArea.startReal) do
					if i > 100 then break end
					local xs,ys,zs = getWorldTranslation(workArea.startReal[i])
					local xw,_,zw = getWorldTranslation(workArea.widthReal[i])
					local xh,_,zh = getWorldTranslation(workArea.heightReal[i])
					workArea.lastStartWorldX[half] = x6
					workArea.lastStartWorldZ[half] = z6
					local ids = g_currentMission.fruits[g_fruitTypeManager.nameToIndex['GRASS']]
					if ids == nil or ids.id == 0 then
						return 0
					end
					local mapModifier = g_currentMission.densityMapModifiers.updateRealMowerArea.modifier
					local mapFilter = g_currentMission.densityMapModifiers.updateRealMowerArea.filter
					local mapMask = g_currentMission.densityMapModifiers.updateRealMowerArea.mask
					mapModifier:setParallelogramWorldCoords( xs,zs, xw,zw, xh,zh, 'ppp')
					mapModifier:resetDensityMapAndChannels(ids.id,0,4)
					mapFilter:resetDensityMapAndChannels(ids.id,0,4)
					mapMask:setValueCompareParams('equal', 0)
					mapFilter:setValueCompareParams('equal', 0)
					local areaValue, modArea, area = mapModifier:executeSet(1,mapMask,mapFilter)
					if modArea ~= 0 then
						workArea.infoText = workArea.infoText .. i.." - "..area.." modArea - "..modArea.."\n"
						changed[i] = true
						workArea.totalArea = workArea.totalArea + area
						workArea.totalModedArea = workArea.totalModedArea + modArea
					else
						workArea.infoText = workArea.infoText .. i.." - "..area.." modArea - "..modArea.."\n"
						changed[i] = false
					end
				end
				workArea.infoText = workArea.infoText .. n.." - totalArea - "..workArea.totalArea..", totalModedArea - "..workArea.totalModedArea..", "..(workArea.totalModedArea*100)/workArea.totalArea.."%\n"
				for i, area in pairs(workArea.startReal) do
					if changed[i] == true then
						workArea.infoText = workArea.infoText .. i.." seed - "
						local xs,ys,zs = getWorldTranslation(workArea.startReal[i])
						local xw,_,zw = getWorldTranslation(workArea.widthReal[i])
						local xh,_,zh = getWorldTranslation(workArea.heightReal[i])
							
						spec.groundRaycastResult.object = 0
						spec.groundRaycastResult.y = ys - 1
						raycastClosest(xs, ys, zs, 0.0, -1.0, 0.0, "groundRaycastCallback", 5, self, 59)
						local hitTerrain = spec.groundRaycastResult.object == g_currentMission.terrainRootNode
						local _, _, _, a , materialId = getTerrainAttributesAtWorldPos(g_currentMission.terrainRootNode, xs, ys, zs, true, true, true, true, false)
						local number = #workArea.startReal
						if (i == 1 and changed[i+1] == true) or (i == number and changed[number-1] == true) or (changed[i-1] == true and changed[i+1] == true) or not hitTerrain or a == 0 then
							FSDensityMapUtil.removeFieldArea(xs,zs, xw,zw, xh,zh)
							workArea.infoText = workArea.infoText .. "removed, "..tostring(hitTerrain)..", "..a..", "..materialId.."\n"
							workArea.totalRemoved = workArea.totalRemoved + 1
						else
							workArea.infoText = workArea.infoText .. "not removed, "..tostring(hitTerrain)..", "..a..", "..materialId.."\n"
							workArea.totalNotRemoved = workArea.totalNotRemoved + 1
						end
					else
						workArea.infoText = workArea.infoText .. i.." not\n"
					end
				end
				workArea.infoText = workArea.infoText .. n.." - totalRemoved - "..workArea.totalRemoved..", totalNotRemoved - "..workArea.totalNotRemoved.."\n----------------------------------------\n"
				workArea.infoTextTemp = workArea.infoText
			end
			if workArea.infoText ~= '' then
				spec.infoText = spec.infoText .. workArea.infoText
			else
				spec.infoText = spec.infoText .. workArea.infoTextTemp
			end
		end
		--renderText(0.7,0.8, 0.017, spec.infoText)
	end
	if self.isClient then
		if self:getIsActiveForInput() then
			if spec.event_IDs ~= nil and g_dedicatedServerInfo == nil then
				for actionName,eventID in pairs(spec.event_IDs) do
					if actionName == InputAction.REALMOWER_ACTIVATE then
						g_inputBinding:setActionEventActive(eventID, true)
						g_inputBinding:setActionEventText(eventID, spec.modActive and g_i18n:getText('REALMOWER_DEACTIVATE') or g_i18n:getText('REALMOWER_ACTIVATE'))
					end
				end
			end
		end
	end
end
function RealMower:actionCallback(actionName, keyStatus, arg4, arg5, arg6)
	local spec = self.spec_realMower
	
	if not spec.modInitialized then
		return
	end
	
	if keyStatus > 0 then
		if actionName == 'REALMOWER_ACTIVATE' then
			spec.modActive = not spec.modActive
			self:toggleActive(spec.modActive, false)
		end
	end
end
function RealMower:toggleActive(modActive, noEventSend)
	local spec = self.spec_realMower
	if not spec.modInitialized then
		return
	end
	
	spec.modActive = modActive
	
	RealMowerToggleActiveEvent.sendEvent(self, modActive, noEventSend)
end
function RealMower:onReadStream(streamId, connection)
	local spec = self.spec_realMower
	if not spec.modInitialized then
		return
	end
	
    local modActive = streamReadBool(streamId)
	if modActive ~= nil then
		self:toggleActive(modActive, true)
	end
end
function RealMower:onWriteStream(streamId, connection)
	local spec = self.spec_realMower
	if not spec.modInitialized then
		return
	end
	
    streamWriteBool(streamId, spec.modActive)
end

RealMowerToggleActiveEvent = {};
RealMowerToggleActiveEvent_mt = Class(RealMowerToggleActiveEvent, Event)

InitEventClass(RealMowerToggleActiveEvent, "RealMowerToggleActiveEvent")

function RealMowerToggleActiveEvent:emptyNew()
    local self = Event:new(RealMowerToggleActiveEvent_mt)
    self.className="RealMowerToggleActiveEvent"
    return self
end

function RealMowerToggleActiveEvent:new(object, modActive)
	local self = RealMowerToggleActiveEvent:emptyNew()
	self.object = object
	self.modActive = modActive
	return self
end

function RealMowerToggleActiveEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
    self.modActive = streamReadBool(streamId)
    self:run(connection)
end

function RealMowerToggleActiveEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.modActive)
end

function RealMowerToggleActiveEvent:run(connection)
	if self.object ~= nil then
		self.object:toggleActive(self.modActive, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(RealMowerToggleActiveEvent:new(self.object, self.modActive), nil, connection, self.object)
	end
end

function RealMowerToggleActiveEvent.sendEvent(vehicle, modActive, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(RealMowerToggleActiveEvent:new(vehicle, modActive), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(RealMowerToggleActiveEvent:new(vehicle, modActive))
		end
	end
end